package com.aliasi.lingmed.mesh;

import java.io.File;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;

import org.xml.sax.SAXException;

import com.aliasi.corpus.ObjectHandler;
import com.aliasi.dict.ApproxDictionaryChunker;
import com.aliasi.dict.DictionaryEntry;
import com.aliasi.dict.ExactDictionaryChunker;
import com.aliasi.dict.MapDictionary;
import com.aliasi.dict.TrieDictionary;
import com.aliasi.spell.FixedWeightEditDistance;
import com.aliasi.spell.WeightedEditDistance;
import com.aliasi.tokenizer.IndoEuropeanTokenizerFactory;
import com.aliasi.tokenizer.TokenizerFactory;
import com.aliasi.util.AbstractExternalizable;

/**
 * The {@code MeshDictionaryCommand} class provides a static main method for
 * compiling a chunker based on MeSH terms.
 * 
 * @author Bob Carpenter
 * @version 1.4
 * @since LingMed1.4
 */
public class MeshDictionaryCommand {

	private MeshDictionaryCommand() {
		/* no instances */
	}

	public static void main(String[] args) throws SAXException, IOException {
		File meshFile = new File(args[0]);
		File dictChunkerFile = new File(args[1]);
		// dictChunkerFile.mkdirs();

		MeshParser parser = new MeshParser();
		DictionaryHandler handler = new DictionaryHandler();
		parser.setHandler(handler);

		parser.parse(meshFile);
		handler.compileDictionary(dictChunkerFile);
	}

	static final TokenizerFactory TOKENIZER_FACTORY = IndoEuropeanTokenizerFactory.INSTANCE;

	static class DictionaryHandler implements ObjectHandler<Mesh> {
		final TrieDictionary<String> mDictionary = new TrieDictionary<String>();

		public void handle(Mesh mesh) {
			MeshNameUi descriptor = mesh.descriptor();
			String headingName = descriptor.name();
			for (MeshConcept concept : mesh.conceptList()) {
				for (MeshTerm term : concept.termList()) {
					MeshNameUi termNameUi = term.nameUi();
					String termName = termNameUi.name();
					DictionaryEntry<String> entry = new DictionaryEntry<String>(termName, headingName);
					mDictionary.addEntry(entry);
				}
			}
		}

		void compileDictionary(File file) throws IOException {
			ChunkerSerializer serializer = new ChunkerSerializer(mDictionary);
			AbstractExternalizable.serializeTo(serializer, file);
		}

	}

	static class ChunkerSerializer extends AbstractExternalizable {
		static final long serialVersionUID = 1879892051066513198L;
		private final TrieDictionary<String> mDictionary;

		public ChunkerSerializer() {
			this(null);
		}

		public ChunkerSerializer(TrieDictionary<String> dictionary) {
			mDictionary = dictionary;
		}

		@Override
		public Object read(ObjectInput in) throws IOException, ClassNotFoundException {
			Object inObject = in.readObject();
			if (inObject instanceof MapDictionary<?>) {
				@SuppressWarnings("unchecked")
				MapDictionary<String> dictionary = (MapDictionary<String>) inObject;
				boolean returnAllMatches = true;
				boolean caseSensitive = false;
				return new ExactDictionaryChunker(dictionary, TOKENIZER_FACTORY, returnAllMatches, caseSensitive);
			} else {
				@SuppressWarnings("unchecked")
				TrieDictionary<String> dictionary = (TrieDictionary<String>) inObject;
				WeightedEditDistance editDistance = new FixedWeightEditDistance(0, -1, -1, -1, Double.NaN);
				double maxDistance = 2.0;
				return new ApproxDictionaryChunker(dictionary, TOKENIZER_FACTORY, editDistance, maxDistance);
			}
		}

		@Override
		public void writeExternal(ObjectOutput out) throws IOException {
			out.writeObject(mDictionary);
		}

	}

}