package no.ntnu.jervidal.master.medline.evaluation;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Collections;
import java.util.List;

import no.ntnu.jervidal.master.Main;
import no.ntnu.jervidal.master.medline.MyMedlineSearcher;
import no.ntnu.jervidal.master.medline.analyzers.bm25.BM25FScorerWithMesh;
import no.ntnu.jervidal.master.medline.results.SearchResult;
import no.ntnu.jervidal.master.medline.results.SearchResults;
import no.ntnu.jervidal.master.medline.searchers.BM25;
import no.ntnu.jervidal.master.medline.searchers.BM25GeneTag;
import no.ntnu.jervidal.master.medline.searchers.BM25GeneTagAndMesh;
import no.ntnu.jervidal.master.medline.searchers.BM25GeneTagGeniaAndMesh;
import no.ntnu.jervidal.master.medline.searchers.BM25Genia;
import no.ntnu.jervidal.master.medline.searchers.BM25GeniaAndMesh;
import no.ntnu.jervidal.master.medline.searchers.BM25Mesh;
import no.ntnu.jervidal.master.medline.searchers.LuceneDefault;
import no.ntnu.jervidal.master.medline.searchers.MedlineSearcher;
import no.ntnu.jervidal.master.medline.searchers.Mesh;
import no.ntnu.jervidal.master.medline.searchers.MeshPhrase;

import org.apache.log4j.Logger;
import org.apache.lucene.queryParser.ParseException;

public class TrecEval {

	/**
	 * <p>
	 * Example result: <code>1 Q0 12474524 1 5567 tag1</code>
	 * </p>
	 * 
	 * <p>
	 * From: {@link http://ir.ohsu.edu/genomics/2003protocol.html} where:
	 * <ul>
	 * <li>The first column is the topic number (1-50).</li>
	 * <li>The second column is the query number within that topic. This is
	 * currently unused and must always be Q0.</li>
	 * <li>The third column is the official PubMedID of the retrieved document.</li>
	 * <li>The fourth column is the rank the document is retrieved</li>
	 * <li>The fifth column shows the score (integer or floating point) that
	 * generated the ranking. This score MUST be in descending (non-increasing)
	 * order. The trec_eval program ranks documents based on the scores, not the
	 * ranks in column 4. If a submitter wants the exact ranking submitted to be
	 * evaluated, then the SCORES must reflect that ranking.</li>
	 * <li>The sixth column is called the "run tag" and must be a unique
	 * identifier across all runs submitted to TREC. Thus, each run tag should
	 * have a part that identifies the group and a part that distinguishes runs
	 * from that group. Tags are restricted to 12 or fewer letters and numbers,
	 * and *NO* punctuation, to facilitate labeling graphs and such with the
	 * tags.</li>
	 * </ul>
	 * </p>
	 */
	private static final String RESULTS_TEMPLATE = "%d\t%s\t%s  \t%d\t%f\t%s\n";
	private static final String RUN_TAG = "jervi";
	private static final String FILENAME_TEMPLATE = "%s_results_%s";
	private Logger logger = Logger.getLogger(this.getClass());
	private MedlineSearcher[] algorithms = {
			new LuceneDefault(), new BM25(), new BM25GeneTag(), new BM25Genia(), new Mesh(), new BM25GeneTagAndMesh(),
			new BM25GeniaAndMesh(), new BM25GeneTagGeniaAndMesh(), new BM25Mesh(),
	// new MeshPhrase(),
	// new LuceneDefaultOnlyAbstract()
	};

	public void runEvaluation() {
		long startTime = System.currentTimeMillis() / 1000L;
		for (MedlineSearcher algorithm : algorithms) {
			MyMedlineSearcher searcher = new MyMedlineSearcher(algorithm);
			try {
				writeToFile(searcher.search(), algorithm, startTime);
				computeMeanAndMedianBoostValues(algorithm);
			} catch (ParseException e) {
				logger.error("The query file contains an invalid query.", e);
			} catch (IOException e) {
				logger.error("A low level IO error occured.", e);
			}
		}

	}

	public void writeToFile(List<SearchResults> searchResults, MedlineSearcher searchAlgorithm, long startTime)
			throws IOException {
		String fileName = String.format(FILENAME_TEMPLATE, searchAlgorithm.getClass().getSimpleName(), startTime);
		String directory = Main.properties.getProperty("resultsDirectory");
		String resultsFile = directory + (directory.endsWith(File.separator) ? "" : File.separator) + fileName;
		BufferedWriter writer = new BufferedWriter(new FileWriter(resultsFile));
		logger.info("Writing results to file \"" + resultsFile + "\".");

		int topic = 1;
		for (SearchResults results : searchResults) {
			for (SearchResult result : results) {
				writer.write(formatSingleResult(topic, result));
			}
			topic++;
		}
		writer.close();
	}

	private String formatSingleResult(int topicNumber, SearchResult result) {
		return String.format(RESULTS_TEMPLATE, topicNumber, "Q0", result.getPMID(), result.getRank(),
				result.getScore(), RUN_TAG);
	}

	private void computeMeanAndMedianBoostValues(MedlineSearcher alg) {
		FileWriter fstream;
		BufferedWriter out;
		List<Float> boosts = BM25FScorerWithMesh.getBoosts();
		if (boosts != null && boosts.size() != 0) {
			try {
				fstream = new FileWriter("boosts_" + alg.getClass().getSimpleName() + ".txt");
				out = new BufferedWriter(fstream);
				float sum = 0f;
				for (float boost : boosts) {
					sum += boost;
					out.write(boost + "\n");
				}
				Collections.sort(boosts);
				float median = boosts.get(boosts.size() / 2);
				out.close();
				float average = sum / boosts.size();
				float low = boosts.get(0);
				float high = boosts.get(boosts.size() - 1);
				logger.debug("Low: " + low + "\n" + "High: " + high + "\n" + "Average: " + average + "\n" + "Median: "
						+ median + "\nTotal number of boosts: " + boosts.size());
				BM25FScorerWithMesh.resetBoosts();
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
	}

}
