package no.ntnu.jervidal.master.medline.indexers;

import java.io.File;
import java.io.IOException;
import java.util.HashSet;

import no.ntnu.jervidal.master.Main;
import no.ntnu.jervidal.master.medline.analyzers.AnalyzerFactory;
import no.ntnu.jervidal.master.medline.analyzers.bm25.BM25FParameters;
import no.ntnu.jervidal.master.medline.analyzers.bm25.BM25Similarity;

import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.lucene.index.CorruptIndexException;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.store.Directory;
import org.apache.lucene.store.FSDirectory;

import com.aliasi.chunk.Chunk;
import com.aliasi.chunk.Chunker;
import com.aliasi.lingmed.parser.MeshHeading;
import com.aliasi.lingmed.parser.Topic;
import com.aliasi.util.AbstractExternalizable;

public class BM25GeneTagAndMeshIndexWriter2 extends BM25FIndexWriter {

	private Chunker chunkerGeneTag;
	private Chunker chunkerMesh;
	private float averageTitleLength;
	private float averageAbstractLength;
	private float averageMeshLength;

	private static int counter = 0;

	@Override
	protected IndexWriter initializeIndexWriter() throws IOException {
		File chunkFileGeneTag = new File(Main.properties.getProperty("GeneTagPath"));
		File chunkFileMesh = new File(Main.properties.getProperty("MeshPath"));
		try {
			chunkerGeneTag = (Chunker) AbstractExternalizable.readObject(chunkFileGeneTag);
			chunkerMesh = (Chunker) AbstractExternalizable.readObject(chunkFileMesh);
		} catch (ClassNotFoundException e) {
			logger.error("Could not load the chunker.", e);
		}
		String dirString = Main.properties.getProperty("indexDirectory");
		Directory dir = FSDirectory.open(new File(dirString
				+ (dirString.endsWith(File.separator) ? "" : File.separator)
				+ Main.properties.getProperty("indexDirectoryBM25GeneTagAndMesh")));
		IndexWriter writer = new IndexWriter(dir, AnalyzerFactory.getBM25MedlineAnalyzer(), true,
				IndexWriter.MaxFieldLength.LIMITED);
		writer.setSimilarity(new BM25Similarity());
		averageTitleLength = Float.parseFloat(Main.properties.getProperty("BM25AverageTitleLength", "10.0"));
		averageAbstractLength = Float.parseFloat(Main.properties.getProperty("BM25AverageAbstractLength", "119.0"));
		averageMeshLength = Float.parseFloat(Main.properties.getProperty("BM25AverageMeshLength", "39.0"));
		return writer;
	}

	@Override
	public void addDocument(String titleText, String abstrText, String pmid, MeshHeading[] meshHeadings)
			throws CorruptIndexException, IOException {
		Document doc = new Document();

		Field titleField = new Field(Main.properties.getProperty("docFieldTitle"), titleText, Field.Store.YES,
				Field.Index.ANALYZED);
		Field pmidField = new Field(Main.properties.getProperty("docFieldPMID"), pmid, Field.Store.YES, Field.Index.NO);
		Field abstractField = new Field(Main.properties.getProperty("docFieldAbstract"), abstrText, Field.Store.NO,
				Field.Index.ANALYZED);
		StringBuilder sb = new StringBuilder();
		for (MeshHeading mesh : meshHeadings) {
			for (Topic topic : mesh.topics()) {
				sb.append(topic.topic() + " ");
			}
		}
		String meshString = sb.toString();
		Field meshField = new Field(Main.properties.getProperty("docFieldMesh"), meshString, Field.Store.YES,
				Field.Index.ANALYZED);

		// Chunking titleChunkGeneTag = chunkerGeneTag.chunk(titleText);
		// Chunking abstractChunkGeneTag = chunkerGeneTag.chunk(abstrText);
		// Chunking meshChunkGeneTag = chunkerGeneTag.chunk(sb.toString());
		// Chunking titleChunkGenia = chunkerGenia.chunk(titleText);
		// Chunking abstractChunkGenia = chunkerGenia.chunk(abstrText);
		// Chunking meshChunkGenia = chunkerGenia.chunk(sb.toString());

		Chunker[] chunkers = new Chunker[] {
				chunkerGeneTag, chunkerMesh
		};

		String[] searchStrings = new String[] {
				titleText, abstrText
		};

		float k1 = BM25FParameters.getK1();
		float b = BM25FParameters.getB();
		int localCounter = 0;

		int titleLength = countWords(titleText);
		int abstractLength = countWords(abstrText);
		int meshLength = countWords(meshString);

		HashSet<String> meshSet = new HashSet<String>();

		for (String searchString : searchStrings) {
			HashSet<String> querySet = new HashSet<String>();
			for (Chunker chunker : chunkers) {
				for (Chunk chunk : (chunker.chunk(searchString).chunkSet())) {
					String match = searchString.substring(chunk.start(), chunk.end());
					querySet.add(match);
					if (meshString.contains(match)) {
						meshSet.add(match.toLowerCase());
					}
					if (meshString.contains(chunk.type())) {
						meshSet.add(chunk.type().toLowerCase());
					}
				}
			}
			if (localCounter == 0) {
				titleField.setBoost(calculateFieldBoost(k1, b, titleLength, averageTitleLength, querySet.size()));
			} else if (localCounter == 1) {
				abstractField.setBoost(calculateFieldBoost(k1, b, abstractLength, averageAbstractLength, querySet
						.size()));
				meshField.setBoost(calculateFieldBoost(k1, b, meshLength, averageMeshLength, meshSet.size()));
			}
			localCounter++;
		}
		// titleField.setBoost(calculateFieldBoost(k1, b, titleLength,
		// averageTitleLength, ((float) titleChunkGeneTag
		// .chunkSet().size() + (float) titleChunkGenia.chunkSet().size()) /
		// 2));
		// abstractField.setBoost(calculateFieldBoost(k1, b, abstractLength,
		// averageAbstractLength,
		// ((float) abstractChunkGeneTag.chunkSet().size() + (float)
		// abstractChunkGenia.chunkSet().size()) / 2));
		// meshField.setBoost(calculateFieldBoost(k1, b, numberOfTopics,
		// averageMeshLength, ((float) meshChunkGeneTag
		// .chunkSet().size() + (float) meshChunkGenia.chunkSet().size()) / 2));

		doc.add(titleField);
		doc.add(pmidField);
		doc.add(abstractField);
		doc.add(meshField);

		indexWriter.addDocument(doc);

		if (++counter % 100 == 0) {
			logger.debug("Added document no. " + counter);
		}
	}

	@Override
	public void close() throws CorruptIndexException, IOException {
		float meshLength = BM25Similarity.getLength(Main.properties.getProperty("docFieldMesh", "mesh"))
				/ getNumberOfDocsInIndex();
		logger.info("Average mesh length: " + meshLength);
		Main.properties.setProperty("BM25AverageMeshLength", meshLength + "");
		super.close();
	}

}
