package no.ntnu.jervidal.master.medline.indexers;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import no.ntnu.jervidal.master.Main;
import no.ntnu.jervidal.master.medline.analyzers.AnalyzerFactory;
import no.ntnu.jervidal.master.medline.analyzers.bm25.BM25FParameters;
import no.ntnu.jervidal.master.medline.analyzers.bm25.BM25Similarity;

import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.lucene.index.CorruptIndexException;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.store.Directory;
import org.apache.lucene.store.FSDirectory;

import com.aliasi.chunk.Chunker;
import com.aliasi.chunk.Chunking;
import com.aliasi.lingmed.parser.MeshHeading;
import com.aliasi.lingmed.parser.Topic;
import com.aliasi.util.AbstractExternalizable;

public class BM25MeshIndexWriter extends BM25FIndexWriter {

	private Chunker chunker;
	private float averageTitleLength;
	private float averageAbstractLength;
	private float averageMeshLength;

	private static List<Float> boosts;

	@Override
	protected IndexWriter initializeIndexWriter() throws IOException {
		File chunkFile = new File(Main.properties.getProperty("MeshPath"));
		try {
			chunker = (Chunker) AbstractExternalizable.readObject(chunkFile);
		} catch (ClassNotFoundException e) {
			logger.error("Could not load the chunker.", e);
		}
		String dirString = Main.properties.getProperty("indexDirectory");
		Directory dir = FSDirectory.open(new File(dirString
				+ (dirString.endsWith(File.separator) ? "" : File.separator)
				+ Main.properties.getProperty("indexDirectoryBM25GeneTagAndMesh")));
		IndexWriter writer = new IndexWriter(dir, AnalyzerFactory.getBM25MedlineAnalyzer(), true,
				IndexWriter.MaxFieldLength.LIMITED);
		writer.setSimilarity(new BM25Similarity());
		averageTitleLength = Float.parseFloat(Main.properties.getProperty("BM25AverageTitleLength", "10.0"));
		averageAbstractLength = Float.parseFloat(Main.properties.getProperty("BM25AverageAbstractLength", "119.0"));
		averageMeshLength = Float.parseFloat(Main.properties.getProperty("BM25AverageMeshLength", "10.0"));
		boosts = new ArrayList<Float>();
		return writer;
	}

	@Override
	public void addDocument(String titleText, String abstrText, String pmid, MeshHeading[] meshHeadings)
			throws CorruptIndexException, IOException {
		Document doc = new Document();

		int titleLength = countWords(titleText);
		int abstractLength = countWords(abstrText);

		Field titleField = new Field(Main.properties.getProperty("docFieldTitle"), titleText, Field.Store.YES,
				Field.Index.ANALYZED);
		Field pmidField = new Field(Main.properties.getProperty("docFieldPMID"), pmid, Field.Store.YES, Field.Index.NO);
		Field abstractField = new Field(Main.properties.getProperty("docFieldAbstract"), abstrText, Field.Store.NO,
				Field.Index.ANALYZED);
		// Field titleLengthField = new
		// Field(Main.properties.getProperty("docTitleLength"), titleLength +
		// "",
		// Field.Store.YES, Field.Index.NO);
		// Field abstractLengthField = new
		// Field(Main.properties.getProperty("docAbstractLength"),
		// abstractLength + "",
		// Field.Store.YES, Field.Index.NO);
		StringBuilder sb = new StringBuilder();
		int numberOfTopics = 0;
		for (MeshHeading mesh : meshHeadings) {
			for (Topic topic : mesh.topics()) {
				sb.append(topic.topic() + " ");
				numberOfTopics++;
			}
		}
		// Field meshLengthField = new
		// Field(Main.properties.getProperty("docMeshLength"), numberOfTopics +
		// "",
		// Field.Store.YES, Field.Index.NO);
		Field meshField = new Field(Main.properties.getProperty("docFieldMesh"), sb.toString(), Field.Store.NO,
				Field.Index.ANALYZED);

		Chunking titleChunk = chunker.chunk(titleText);
		Chunking abstractChunk = chunker.chunk(abstrText);
		Chunking meshChunk = chunker.chunk(sb.toString());

		float k1 = BM25FParameters.getK1();
		float b = BM25FParameters.getB();
		float titleBoost = calculateFieldBoost(k1, b, titleLength, averageTitleLength, titleChunk.chunkSet().size());
		float abstractBoost = calculateFieldBoost(k1, b, abstractLength, averageAbstractLength, abstractChunk
				.chunkSet().size());
		float meshBoost = calculateFieldBoost(k1, b, numberOfTopics, averageMeshLength, meshChunk.chunkSet().size());
		titleField.setBoost(titleBoost);
		titleField.setBoost(titleBoost);
		abstractField.setBoost(abstractBoost);
		meshField.setBoost(meshBoost);

		boosts.add((titleBoost + abstractBoost + meshBoost) / 3);

		doc.add(titleField);
		doc.add(pmidField);
		doc.add(abstractField);
		// doc.add(titleLengthField);
		// doc.add(abstractLengthField);
		doc.add(meshField);
		// doc.add(meshLengthField);

		indexWriter.addDocument(doc);
	}

	@Override
	public void close() throws CorruptIndexException, IOException {
		float meshLength = BM25Similarity.getLength(Main.properties.getProperty("docFieldMesh", "mesh"))
				/ getNumberOfDocsInIndex();
		logger.info("Average mesh length: " + meshLength);
		Main.properties.setProperty("BM25AverageMeshLength", meshLength + "");
		computeMeanAndMedianBoostValues();
		super.close();
	}

	private static void resetBoosts() {
		System.out.println("Reseting boosts...");
		boosts.clear();
	}

	private void computeMeanAndMedianBoostValues() {
		FileWriter fstream;
		BufferedWriter out;
		if (boosts != null) {
			try {
				fstream = new FileWriter("boosts_" + this.getClass().getSimpleName() + ".txt");
				out = new BufferedWriter(fstream);
				float sum = 0f;
				for (float boost : boosts) {
					sum += boost;
					out.write(boost + "\n");
				}
				Collections.sort(boosts);
				float median = boosts.get(boosts.size() / 2);
				out.close();
				float average = sum / boosts.size();
				float low = boosts.get(0);
				float high = boosts.get(boosts.size() - 1);
				logger.debug("Low: " + low + "\n" + "High: " + high + "\n" + "Average: " + average + "\n" + "Median: "
						+ median + "\nTotal number of boosts: " + boosts.size());
				resetBoosts();
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
	}

}
