package no.ntnu.jervidal.master.medline.indexers;

import java.io.IOException;

import org.apache.log4j.Logger;
import org.apache.lucene.index.CorruptIndexException;
import org.apache.lucene.index.IndexWriter;

import com.aliasi.lingmed.parser.MeshHeading;

public abstract class MedlineIndexWriter {
	protected Logger logger;
	protected IndexWriter indexWriter;

	public MedlineIndexWriter() {
		logger = Logger.getLogger(this.getClass());
		try {
			indexWriter = initializeIndexWriter();
		} catch (IOException e) {
			logger.error("Could not initialize the index.", e);
		}
	}

	/**
	 * This method should instantiate and return an {@link IndexWriter} object
	 * that will be used during the indexing. The method will be called from the
	 * constructor when instantiating the class, so if overriding the
	 * constructor, be sure to either call <code>super()</code> or call the
	 * method yourself.
	 * 
	 * @throws IOException
	 */
	protected abstract IndexWriter initializeIndexWriter() throws IOException;

	/**
	 * This method adds a new document to the index.
	 * 
	 * @param titleText
	 *            The title of the document
	 * @param abstrText
	 *            The abstract of the document
	 * @param pmid
	 *            The PMID (PubMed ID) of the document
	 * @param meshHeadings
	 *            The {@link MeshHeading}s of the document in an array
	 * @throws CorruptIndexException
	 *             Thrown if the index is corrupt.
	 * @throws IOException
	 *             Thrown if an IOException occurs.
	 */
	public abstract void addDocument(String titleText, String abstrText, String pmid, MeshHeading[] meshHeadings)
			throws CorruptIndexException, IOException;

	/**
	 * <p>
	 * <strong>Description copied from {@link IndexWriter#close()}:</strong>
	 * </p>
	 * <p>
	 * Commits all changes to an index and closes all associated files. Note
	 * that this may be a costly operation, so, try to re-use a single writer
	 * instead of closing and opening a new one.
	 * </p>
	 * 
	 * @throws CorruptIndexException
	 * @throws IOException
	 */
	public void close() throws CorruptIndexException, IOException {
		indexWriter.close();
	}

	/**
	 * <p>
	 * <strong>Description copied from {@link IndexWriter#numDocs()}:</strong>
	 * </p>
	 * <p>
	 * Returns total number of docs in this index, including docs not yet
	 * flushed (still in the RAM buffer), and including deletions.
	 * <strong>NOTE:</strong> buffered deletions are not counted.
	 * </p>
	 * 
	 * @return The total number of docs in this index, or -1 if any errors
	 *         occured.
	 */
	public int getNumberOfDocsInIndex() {
		try {
			return indexWriter.numDocs();
		} catch (IOException e) {
			logger.warn("Could not calculate the number of documents in the index.", e);
			return -1;
		}
	}

	/**
	 * <p>
	 * <strong>Description copied from {@link IndexWriter#optimize()}:</strong>
	 * </p>
	 * 
	 * <p>
	 * Requests an "optimize" operation on an index, priming the index for the
	 * fastest available search. Traditionally this has meant merging all
	 * segments into a single segment as is done in the default merge policy,
	 * but individual merge policies may implement optimize in different ways.
	 * It is recommended that this method be called upon completion of indexing.
	 * In environments with frequent updates, optimize is best done during low
	 * volume times, if at all.
	 * </p>
	 * 
	 * <p>
	 * See {@link http://www.gossamer-threads.com/lists/lucene/java-dev/47895}
	 * for more discussion.
	 * </p>
	 * 
	 * <p>
	 * Note that optimize requires 2X the index size free space in your
	 * Directory. For example, if your index size is 10 MB then you need 20 MB
	 * free for optimize to complete.
	 * </p>
	 * 
	 * @throws CorruptIndexException
	 * @throws IOException
	 */
	public void optimize() throws CorruptIndexException, IOException {
		indexWriter.optimize();
	}
}
