package no.ntnu.jervidal.master.medline.searchers;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import no.ntnu.jervidal.master.Main;
import no.ntnu.jervidal.master.medline.analyzers.AnalyzerFactory;
import no.ntnu.jervidal.master.medline.analyzers.bm25.BM25FParameters;
import no.ntnu.jervidal.master.medline.analyzers.bm25.BM25Similarity;
import no.ntnu.jervidal.master.medline.analyzers.bm25.BM25TermQuery;

import org.apache.lucene.index.CorruptIndexException;
import org.apache.lucene.queryParser.MultiFieldQueryParser;
import org.apache.lucene.queryParser.ParseException;
import org.apache.lucene.queryParser.QueryParser;
import org.apache.lucene.search.BooleanClause;
import org.apache.lucene.search.BooleanQuery;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.search.Query;
import org.apache.lucene.search.TermQuery;
import org.apache.lucene.store.FSDirectory;
import org.apache.lucene.util.Version;

public class BM25 extends MedlineSearcher {

	protected QueryParser parser;
	protected String[] fields;
	protected Map<String, Float> boosts;

	@Override
	public IndexSearcher initializeIndexSearcher() throws CorruptIndexException, IOException {
		String dirString = Main.properties.getProperty("indexDirectory");
		IndexSearcher searcher = new IndexSearcher(FSDirectory.open(new File(dirString
				+ (dirString.endsWith(File.separator) ? "" : File.separator)
				+ Main.properties.getProperty("indexDirectoryBM25"))));
		BM25FParameters.setAverageLength(Main.properties.getProperty("docFieldTitle"), Float.parseFloat(Main.properties
				.getProperty("BM25AverageTitleLength")));
		BM25FParameters.setAverageLength(Main.properties.getProperty("docFieldAbstract"), Float
				.parseFloat(Main.properties.getProperty("BM25AverageAbstractLength")));
		BM25FParameters.setK1(2.0f);
		searcher.setSimilarity(new BM25Similarity());
		fields = new String[] {
				Main.properties.getProperty("docFieldTitle", "title"),
				Main.properties.getProperty("docFieldAbstract", "abstract")
		};
		float[] bParams = new float[fields.length];
		for (int i = 0; i < fields.length; i++) {
			bParams[i] = 0.75f;
		}
		BM25FParameters.setBParam(bParams);
		boosts = new HashMap<String, Float>();
		boosts.put(fields[0], Float.parseFloat(Main.properties.getProperty("titleBoost")));
		boosts.put(fields[1], Float.parseFloat(Main.properties.getProperty("abstractBoost")));
		return searcher;
	}

	@Override
	public Query parseQuery(String query) throws ParseException {
		if (parser == null) {
			parser = new MultiFieldQueryParser(Version.LUCENE_30, fields, AnalyzerFactory.getBM25MedlineAnalyzer(),
					boosts);
		}
		Query parsedQuery = parser.parse(query);
		return replaceQuery(parsedQuery);
	}

	private Query replaceQuery(Query q) throws ParseException {
		if (q instanceof TermQuery) {
			TermQuery tq = (TermQuery) q;
			BM25TermQuery bm25tq;
			if (fields == null)
				bm25tq = new BM25TermQuery(tq.getTerm());
			else {
				float[] boostsArray = new float[boosts.size()];
				for (int i = 0; i < boostsArray.length; i++) {
					boostsArray[i] = boosts.get(fields[i]);
				}
				bm25tq = new BM25TermQuery(tq.getTerm(), fields, boostsArray, BM25FParameters.getBParam());
			}
			bm25tq.setBoost(tq.getBoost());
			return bm25tq;
		} else if (q instanceof BooleanQuery) {
			replaceBooleanTermQuery((BooleanQuery) q);
			return q;
			// } else if (q instanceof PhraseQuery) {
			// PhraseQuery pq = (PhraseQuery)q;
			// logger.debug("Phrase query: " + pq.toString());
			// BooleanQuery bq = new BooleanQuery()
			// BM25TermQuery bm25tq;
			// if (fields == null)
			// bm25tq = new BM25TermQuery();
			// else {
			// float[] boostsArray = new float[boosts.size()];
			// for (int i = 0; i < boostsArray.length; i++) {
			// boostsArray[i] = boosts.get(fields[i]);
			// }
			// bm25tq = new BM25TermQuery(tq.getTerm(), fields, boostsArray,
			// BM25FParameters.getBParam());
			// }
			// bm25tq.setBoost(tq.getBoost());
			// return bm25tq;
		} else {
			throw new ParseException("unsupported query type " + q.getClass() + ": " + q.toString());
		}
	}

	private void replaceBooleanTermQuery(BooleanQuery boolQuery) throws ParseException {
		for (BooleanClause clause : boolQuery.getClauses()) {
			Query q = clause.getQuery();
			Query newQ = replaceQuery(q);
			clause.setQuery(newQ);
		}
	}
}
